#amarok Plus implementation which overrides amarok proxy regular flavor with extended functionality
# Manuel Amador (Rudd-O) - written on 2005-11-28


class NotAnalyzedYet(Exception): pass
class UnAnalyzable(Exception): pass


from amarok import amaroKProxy,CollectionSong,amaroKCollection,DcopAmarokFailed,DBTYPE_OTHER,DBTYPE_MYSQL
from songanalysiswrapper import get_interface
from songanalysiscommon import *

class CollectionSongPlus(CollectionSong):
	
	def __get_tempo(self):
		if not hasattr(self,"_tempo"): self._tempo = self.collectionObject._getTempo(self.path)
		return self._tempo
	
	def __set_tempo(self,tempo):
		self.collectionObject._setTempo(self.path,tempo)
		self._tempo = tempo
	
	tempo = property(__get_tempo,__set_tempo)
	
	
class amaroKCollectionPlus(amaroKCollection):
	
	def getSong(self,path): # we override to provide new functionality
		return CollectionSongPlus(path,self)
	
	def _getTempo(self,path):
		try:
			tempo = self.queryOne("select bpm from analysis where analysis.url = %s;"%self.sql_escape(path))
			if tempo is None: return None
			return float(tempo)
		except ValueError,e: return None
		
	def _setTempo(self,path,tempo): # the extattr shit should be done here and in the preceding function
		if self.query("select url from analysis where analysis.url = %s;"%self.sql_escape(path)):
			self.query("update analysis set bpm = %f where analysis.url = %s;"%(tempo,self.sql_escape(path)))
		else:
			self.query("insert into analysis (url,bpm,interface_used) values(%s,%f,%s);"%(
					self.sql_escape(path),tempo,self.sql_escape(get_interface())))

	def __init__(self,parent):
		amaroKCollection.__init__(self,parent)
		self.__prepareDatabase()
		
	def __prepareDatabase(self):
		self.query(n_create_sql)
		self.purgeObsoleteAnalyses()
		
	def purgeObsoleteAnalyses(self):
		sql1 = "delete from analysis where interface_used is NULL;"
		sql2 = "delete from analysis where interface_used != %s;"%self.sql_escape(get_interface())
		for a in [sql1,sql2]: self.query(a)
		
	def getNextUnanalyzedSong(self):
		sql = "select tags.url from tags left join analysis on (tags.url = analysis.url) where analysis.url is NULL limit 1;"
		return self.queryOne(sql)

	def countUnanalyzedSongs(self):
		sql = "select count(tags.url) as remaining from tags left join analysis on (tags.url = analysis.url) where analysis.url is NULL;"
		return int(self.queryOne(sql))

	def _forgetAnalysisData(self,filename):
		self.query("delete from analysis where url = %s;"%self.sql_escape(filename))
	
	def associate(self,filename,tempo,freqbands,voldiff,interface_used): # the extattr shit should be done here
		
		def is_infinite(value):
			if value == float("inf"): return True
			if value == float("-inf"): return True
			if str(value) == "nan": return True
			return False
			
		def sqlize(value):
			if value is None or is_infinite(value): return "NULL"
			return "%f"%value
			
		tempo = sqlize(tempo)
		voldiff = sqlize(voldiff)
		if freqbands is None: freqbands = ",".join([ "NULL" for r in range(30) ])
		else: freqbands = ",".join([ sqlize(f) for f in freqbands ])
		fnames = ",".join([ "freq%s"%s for s in range(30) ])
		
		sql = "insert into analysis (url,volume_diff,%s,bpm,interface_used) values(%s,%s,%s,%s,%s);" %(fnames,self.sql_escape(filename),voldiff,freqbands,tempo,self.sql_escape(interface_used))
		
		self._forgetAnalysisData(filename)
		self.query(sql)
		
		
	def getSimilarSongs(self,filename,weight_freq,weight_bpm,limit,fixed_tempo=None,subquery=None):
		
		# FIXME revisar este algorithmo
		if self.databaseType is DBTYPE_OTHER:
			t_max="MAX"
			t_min="MIN"
		elif self.databaseType is DBTYPE_MYSQL:
			t_max="GREATEST"
			t_min="LEAST"
		else: assert(False)
		
		
		output = self.query("select * from analysis where url = %s;"% self.sql_escape(filename))
		if not output: raise NotAnalyzedYet,filename
		if len(output) != 34: raise UnAnalyzable,"%s did not produce 34 output fields"%filename
		
		try:
			url,volume_diff,freqbands,bpm = output[0],float(output[1]),[ float(a) for a in output[2:32] ],float(output[32])
		except ValueError, e:
			raise UnAnalyzable,"some stored parameters in %s produced an error: %s"%(filename,output)

		def gen_volume_diff_sql (volume_diff):
			# return the relative difference in volume (positive means louder than the passed volume, negative means quieter)
			if volume_diff is None: volume_diff = "NULL"
			template = "volume_diff - %s"
			sql = template % (volume_diff)
			# we return unadjusted values
			return sql
		def gen_norm_volume_diff_sql (volume_diff):
			# normalized volume value, -1 ... 1, 1 has most similar volume
			s = gen_volume_diff_sql(volume_diff)
			s = t_max + "( -1 , 1 - ABS( %s ) )"%s
			return s
		def gen_band_interference_sql(freqbands):
			# return the magnitude of the compound interference between frequency bands.  the larger the value, the most interference, the less compatible to the freqbands val passed.
		
			portions = []
			freq = freqbands
			
			band = 0
			template = "abs(%s - freq%s) + abs(%s - freq%s)"
			portions.append(template%(freq[band],band,freq[band],band+1))
			
			for band in range(1,29):
				template = "abs(%s - freq%s) + abs(%s - freq%s) + abs(%s - freq%s)"
				portions.append(template%(freq[band],band,freq[band],band-1,freq[band],band+1))
			
			band = 29
			template = "abs(%s - freq%s) + abs(%s - freq%s)"
			portions.append(template%(freq[band],band,freq[band],band-1))
			
			sql = " + ".join(portions)
			# 	sql = "1 - (( %s ) / 2.5 )"%sql
			# we return unadjusted values
			
			return sql
		def gen_norm_band_interference_sql(freqbands):
	# 		normalized interference value: -1...1, 1 has most similar freq spectrum distribution
			s = gen_band_interference_sql(freqbands)
			s = "1 - ( ( %s ) / 2.5)" % s
			s = "%s( %s( %s , -1 ) , 1 )" %(t_max,t_min,s)
			return s
		def gen_compatibility_sql(volume_diff,freqbands):
			# this returns -1 ... 1 ranged values which take norm vol and norm interf. and sum them in a weighted sum 3/4 for the interference and 1/4 for the vol difference
			f = gen_norm_band_interference_sql(freqbands)
			v = gen_norm_volume_diff_sql(volume_diff)
			s = "0.75 * ( %s ) + 0.25 * ( %s )"%(f,v)
			return s
		def gen_bpm_diff_sql(bpm):
			if bpm is None: bpm = "NULL"
			sql = "bpm - %s"%bpm
			return sql
		def gen_norm_bpm_diff_sql(bpm):
			s = gen_bpm_diff_sql(bpm)
			s = "ABS( %s ) / (160 - 120)"%s  #MAX_BPM MIN_BPM ... how the fuck do they choose it?
			s = "1 - ( ( %s ) * 2 )"%s
			return s
		def gen_attraction_sql(volume_diff,freqbands,bpm,weight_freq,weight_bpm):
			c = gen_compatibility_sql(volume_diff,freqbands)
			b = gen_norm_bpm_diff_sql(bpm)
			s = "COALESCE( %s * %s , 0 ) + COALESCE( %s * %s , 0 )"%(c,weight_freq,b,weight_bpm)
			# we coalesce so if any value comes null, we simply sum zero to it
			return s
		def gen_force_sql(volume_diff,freqbands,bpm,weight_freq,weight_bpm):
			s = gen_attraction_sql(volume_diff,freqbands,bpm,weight_freq,weight_bpm)
			s = "ABS( %s ) * ( %s )"%(s,s) #we want to keep the sign that's why there is only one abs
			return s
		
		sql = "select COALESCE( tags.url , 'ERROR' ) as url , COALESCE( %s , -1 ) as force_value, COALESCE( bpm , 0) as tempo from analysis inner join tags on analysis.url = tags.url where volume_diff is not NULL %s order by force_value desc limit %s;"
		
		if fixed_tempo is not None: bpm = fixed_tempo
		if not subquery: subquery = ""
		else: subquery = "and tags.url in ( %s )"%subquery
		
		sql = sql%(gen_force_sql(volume_diff,freqbands,bpm,weight_freq,weight_bpm),subquery,limit)
		
		output = self.query(sql)
		assert ( len(output) / 3 ) * 3 == len(output)
		urls = [] ; forces = [] ; bpms = []
		while output:
			a = output.pop(0)
			urls.append(a)
			a = output.pop(0)
			forces.append(a)
			a = output.pop(0)
			bpms.append(a)
		results = []
		while urls and forces and bpms:
			results.append (    [    urls.pop(0) , float(forces.pop(0)) ,    float(bpms.pop(0))  ]      )
		return results


class amaroKProxyPlus(amaroKProxy):
	def __init__(self):
		amaroKProxy.__init__(self)
		self.collection = amaroKCollectionPlus(self)
